import tensorflow as tf
import os
import sys
import numpy as np
import logging
assert sys.version_info[0] == 3

import poeem.ops.python.clustering as clustering
import poeem.ops.python.encode as encode


def write_index_file(index_file, codebook, item_id, item_norm, item_code, centroid, assignment, use_residual=True):
    """Write serving related data into a tsv file in the following format
    use_residual               (int8)
    n_batch d n_cluster K D    (int64)
    codebook                   (float32)
    item_id                    (int64)
    item_norm                  (float32)
    item_code                  (uint8 or uint16)
    centroid                   (float32)
    assignment.indices[:, 1]   (uint32)
    """

    with open(index_file, 'wb') as fp:
        use_residual = np.int8(use_residual)
        use_residual.tofile(fp)

        n_batch, D = np.shape(item_code)
        n_cluster, d = np.shape(centroid)
        K = np.shape(codebook)[1]

        dimensions = np.array([n_batch, d, n_cluster, K, D], dtype=np.int64)
        dimensions.tofile(fp)

        codebook.astype(np.float32).tofile(fp)
        item_id.astype(np.int64).tofile(fp)
        item_norm.astype(np.float32).tofile(fp)
        if K <= 256:
            item_code.astype(np.uint8).tofile(fp)
        elif K <= 2**16:  # assume K is at most 2^16
            item_code.astype(np.uint16).tofile(fp)
        else:
            raise ValueError("too large value of K")
        centroid.astype(np.float32).tofile(fp)
        assignment.astype(np.uint32).tofile(fp)


def indexing(data_file, index_file, n_cluster, sample_size, max_iter, change_percentage_thr, use_residual=True):
    """Do indexing for input data and a codebook

    Args:
        data_file: a npz file of a tensor with item_id in [n_batch], item_code in [n_batch, D], codebook in [D, K, d/D]
        K: a scalar indicating number of clusters for coarse quantization
        index_file: output index file

    Returns:
        None
    """
    with open(data_file, 'rb') as fp:
        all_data = np.load(fp, allow_pickle=True)
        item_code = all_data['item_code']
        item_id = all_data['item_id']
        item_norm = all_data['item_norm']
        codebook = all_data['codebook']

    if use_residual:
        # thus, no need to do clustering
        with open(data_file, 'rb') as fp:
            all_data = np.load(fp, allow_pickle=True)
            centroid_val = all_data['coarse_codebook']
            assignment_val = all_data['item_coarse_code']
    else:
        item_code_plh = tf.placeholder(tf.uint16, shape=(None, None), name='item_code_plh')
        codebook_plh = tf.placeholder(tf.float32, shape=(None, None, None), name='codebook_plh')
        centroid, assignment = clustering.kmeans(
            item_code_plh, codebook_plh, n_cluster, sample_size, max_iter, change_percentage_thr)
        with tf.Session() as sess:
            centroid_val, assignment_val = sess.run([centroid, assignment], feed_dict={
                item_code_plh: item_code,
                codebook_plh: codebook
            })
        
    # write the index file
    write_index_file(index_file, codebook, item_id, item_norm, item_code, 
                     centroid_val, assignment_val, use_residual)

    return None




__all__ = ["clustering_op", "clustering_raw_op", "combinatorial_op", "pqknn_dataset_op", "rec_dataset_op", "search_op", "tokenizer_op"]

import tensorflow as tf
import os


module_dir = os.path.join(os.path.dirname(os.path.abspath(__file__)))

clustering_op = tf.load_op_library(os.path.join(module_dir,'clustering_op.so'))
clustering_raw_op = tf.load_op_library(os.path.join(module_dir,'clustering_raw_op.so'))
combinatorial_op = tf.load_op_library(os.path.join(module_dir,'combinatorial_op.so'))
pqknn_dataset_op = tf.load_op_library(os.path.join(module_dir,'pqknn_dataset_op.so'))
rec_dataset_op = tf.load_op_library(os.path.join(module_dir,'rec_dataset_op.so'))
search_op = tf.load_op_library(os.path.join(module_dir,'search_op.so'))
tokenizer_op = tf.load_op_library(os.path.join(module_dir,'tokenizer_op.so'))





import tensorflow as tf
from tensorflow.estimator import ModeKeys
import contextlib

from poeem.ops.python import clustering


def compute_distortion(x, x_tau):
    # x, x_tau : [n_batch, d]
    return tf.reduce_mean(tf.sqrt(tf.reduce_mean(tf.square(x - x_tau), axis=1)))


def compute_rotation(x, x_tau):
    """
    solve the orthogonal Procrustes problem via closed form solution
    check the paper "Optimized Product Quantization for Approximate Nearest Neighbor Search"
    basically,
    min_R ||RX - Y||_F^2
    s.t.   R^T R = I

    the solution to R can be obtained as follows.
    XY^T = USV^T
    R = VU^T

    The below code is for transposed X and Y

    min_R ||XR - Y||_F^2
    s.t.   R^T R = I

    the solution to R can be obtained as follows.
    X^T Y = USV^T
    R = U V^T
    """
    M = tf.matmul(x, x_tau, transpose_a=True)
    s, u, v = tf.svd(M)
    R = tf.matmul(u, v, transpose_b=True)
    return R


class PoeemQuantizer(object):
  def __init__(self, coarse_K, K, D, d, rotate, buffer_size, svd_steps=200):
    """
    Args:
      coarse_K: int, size of coarse quantization
      K, D: int, size of KD code.
      d: dim of continuous input for each of D axis.
      rotate: int, if we want to learn a rotation to minimize distortion
      buffer_size: data buffer to initialize centroids
    """
    self._coarse_K = coarse_K
    self._K = K
    self._D = D
    self._d = d
    self._sub_d = d // D
    self._buffer_size = buffer_size
    self._rotate = rotate
    self._svd_steps = svd_steps
    assert self._sub_d >= 1, '_sub_d = 0, _sub_d should be greater and equal than 1, this is caused by that D is greater than embedding dimension'

    self._sample_buffer = tf.Variable(
      name='sample_buffer',
      initial_value=tf.zeros([0, self._d]),
      shape=[None, self._d],
      use_resource=True,  # necessary to have undetermined dimension size.
      trainable=False,
      aggregation=tf.VariableAggregation.SUM)
    self._sample_size = tf.get_variable(
      name='sample_size',
      shape=[],
      dtype=tf.int32,
      initializer=tf.zeros_initializer(),
      use_resource=True,  # necessary if variable is used in tf.cond condition
      trainable=False,
      aggregation=tf.VariableAggregation.SUM)
    self._initialized = tf.get_variable(
      name='initialized',
      shape=[],
      dtype=tf.int32,
      initializer=tf.zeros_initializer(),
      use_resource=True,  # necessary if variable is used in tf.cond condition
      trainable=False,
      aggregation=tf.VariableAggregation.SUM)
    self._coarse_centroids = tf.get_variable(
      name='coarse_centroids',
      shape=[self._coarse_K, self._d],
      aggregation=tf.VariableAggregation.SUM) if self._coarse_K > 0 else None
    self._centroids = tf.get_variable(
      name="centroids_k",
      shape=[self._D, self._K, self._sub_d],
      aggregation=tf.VariableAggregation.SUM)
    self._rotation_matrix = tf.get_variable(
      name="rotation_matrix",
      shape=[self._d, self._d],
      initializer=tf.keras.initializers.Identity(),
      trainable=True,
      aggregation=tf.VariableAggregation.SUM) if rotate else None

  def coarse_l2_distance(self, x, coarse_centroids):
    n_batch = tf.shape(x)[0]
    norm_1 = tf.tile(
        tf.reduce_sum(x**2, axis=-1, keep_dims=True),  # n_batch x coarse_K
        [1, self._coarse_K])
    norm_2 = tf.tile(
        tf.transpose(
            tf.reduce_sum(coarse_centroids**2, axis=-1, keep_dims=True),
            [1, 0]),                                   # n_batch x coarse_K
        [n_batch, 1])
    # x = tf.Print(x, [tf.shape(x), tf.shape(coarse_centroids)], message='before dot ')
    dot = tf.matmul(x, coarse_centroids, transpose_b=True) # n_batch x coarse_K
    # dot = tf.Print(dot, [tf.shape(x), tf.shape(coarse_centroids), tf.shape(dot)], message='after dot ')
    # norm_1 = tf.Print(norm_1, [tf.shape(norm_1), tf.shape(norm_2), tf.shape(dot)], message='before l2_sqr ')
    l2_sqr = norm_1 + norm_2 - 2 * dot                 # n_batch x coarse_K
    # l2_sqr = tf.Print(l2_sqr, [tf.shape(norm_1), tf.shape(norm_2), tf.shape(dot)], message='after l2_sqr ')
    return l2_sqr

  def pq_l2_distance(self, x, centroids):
    n_batch = tf.shape(x)[0]
    x = tf.reshape(x, [n_batch, self._D, self._sub_d])
    norm_1 = tf.reduce_sum(x**2, -1, keep_dims=True)        # (n_batch, D, 1)
    norm_2 = tf.expand_dims(tf.reduce_sum(centroids**2, -1), 0)                         # (1, D, K)
    dot = tf.matmul(tf.transpose(x, perm=[1, 0, 2]),
                    tf.transpose(centroids, perm=[0, 2, 1]))    # (D, n_batch, K)
    l2_sqr = norm_1 + norm_2 - 2 * tf.transpose(dot, perm=[1, 0, 2])  # (n_batch, D, K)
    return l2_sqr

  def quantize(self, x, coarse_centroids, centroids):
    # x: (batch_size, d)
    n_batch = tf.shape(x)[0]
    coarse_code, coarse_output = None, None

    if self._coarse_K > 0:
        # coarse_K > 0 means use residual
        coarse_l2_sqr = self.coarse_l2_distance(x, coarse_centroids)
        coarse_code = tf.argmin(coarse_l2_sqr, -1)
        coarse_output = tf.nn.embedding_lookup(
            coarse_centroids, coarse_code)

        # compute residual
        x = x - coarse_output

    l2_sqr = self.pq_l2_distance(x, centroids)      # (n_batch, D, K)
    code = tf.argmin(l2_sqr, -1)                    # (n_batch, D)
    neighbor_idxs = code

    # Compute the outputs, which has shape (batch_size, D, sub_d)
    D_base = tf.convert_to_tensor(
        [self._K*d for d in range(self._D)], dtype=tf.int64)
    neighbor_idxs += tf.expand_dims(D_base, 0)       # (batch_size, D)
    neighbor_idxs = tf.reshape(neighbor_idxs, [-1])  # (batch_size * D)
    centroids = tf.reshape(centroids, [-1, self._sub_d])
    output = tf.nn.embedding_lookup(centroids, neighbor_idxs)
    output = tf.reshape(output, [n_batch, self._d])

    if self._coarse_K > 0:
      x_tau = coarse_output + output
    else:
      x_tau = output

    return x_tau, coarse_code, code

  def compute_centroids(self, x, max_iter=100, change_pct_thr=0.01):
    coarse_centroids = None
    # run coarse clustering and clustering in parallel
    if self._coarse_K > 0:
      coarse_centroids, _ = clustering.kmeans_raw(
        x, self._coarse_K,
        max_iter=max_iter, change_percentage_thr=change_pct_thr, verbose=1)
      coarse_centroids = tf.reshape(coarse_centroids, [self._coarse_K, self._d])

    centroids_list = []
    for i in range(self._D):
      ctrd, _ = clustering.kmeans_raw(
        x[:, i*self._sub_d:(i+1)*self._sub_d], self._K,
        max_iter=max_iter, change_percentage_thr=change_pct_thr, verbose=1)
      centroids_list.append(ctrd)
    centroids = tf.stack(centroids_list, axis=0)
    centroids = tf.reshape(centroids, [self._D, self._K, self._sub_d])
    return coarse_centroids, centroids

  def forward(self, x, rotation_matrix=None):
    """Rotate x, compute quantized  x_tau, then rotate x_tau back."""
    with tf.name_scope("Poeem_forward"):
      if self._rotate > 0:
        assert rotation_matrix is not None
        x = tf.matmul(x, rotation_matrix)
        x_tau, coarse_code, code = self.quantize(
            x, self._coarse_centroids, self._centroids)
        x_tau = tf.matmul(x_tau, tf.transpose(rotation_matrix, [1, 0]))
      else:
        x_tau, coarse_code, code = self.quantize(
            x, self._coarse_centroids, self._centroids)

      return x_tau, coarse_code, code

  def accumulate(self, x):
    # cutoff data to fit in sample buffer
    batch_size = tf.minimum(
      tf.shape(x)[0],
      self._buffer_size - self._sample_size)
    x_sub = tf.slice(x, [0, 0], [batch_size, -1])

    assign_tensor = self._sample_buffer.assign(
      tf.concat([self._sample_buffer, x_sub], axis=0))

    sample_size_tensor = self._sample_size.assign_add(batch_size)
    return assign_tensor, sample_size_tensor

  def init_centroids(self, max_iter=None, change_pct_thr=0.01):
    assign_ops = []
    if self._rotate > 0:
      def loop_body(i, R, data):
        x = tf.matmul(data, R)
        coarse_centroids, centroids = self.compute_centroids(
          x, max_iter=max_iter or 10, change_pct_thr=change_pct_thr)
        x_tau, coarse_code, code = self.quantize(x, coarse_centroids, centroids)
        distortion = compute_distortion(x, x_tau)
        x_tau = tf.Print(x_tau, [i, distortion], message='distortion = ')
        R = compute_rotation(data, x_tau)
        i = tf.add(i, 1)
        return i, R, data

      def condition(i, R, data):
        return tf.less(i, self._svd_steps)

      # find the optimized rotation to minimize distortion, by alternative minimization
      i = tf.constant(0)
      R = tf.eye(self._d)
      data = tf.reshape(self._sample_buffer, [self._buffer_size , self._d])  # necessary to avoid dimension inconsistency error
      # run while_loop in sequence
      _, R, _ = tf.while_loop(
        condition, loop_body, [i, R, data],
        shape_invariants=[
          tf.TensorShape([]),
          tf.TensorShape([self._d, self._d]),
          tf.TensorShape([self._buffer_size, self._d])])

      # apply the rotation
      x = tf.matmul(self._sample_buffer, R)
      coarse_centroids, centroids = self.compute_centroids(
        x, max_iter=max_iter or 100, change_pct_thr=change_pct_thr)
      assign_ops.append(self._rotation_matrix.assign(R))
    else:
      coarse_centroids, centroids = self.compute_centroids(self._sample_buffer,
        max_iter=max_iter or 100, change_pct_thr=change_pct_thr)
    if self._coarse_K > 0:
      assign_ops.append(self._coarse_centroids.assign(coarse_centroids))
    assign_ops.append(self._centroids.assign(centroids))
    assign_ops.append(self._initialized.assign(1))
    return assign_ops

  def clear_sample_buffer(self):
    buffer_assign = tf.assign(self._sample_buffer, tf.zeros([0, self._d]), validate_shape=False)
    return buffer_assign,


class PoeemEmbed(object):
  def __init__(self, emb_size, warmup_steps=1024, buffer_size=8192, mode=ModeKeys.PREDICT,
      hparams=None, name="poeem_emb", initializer=None, svd_steps=200):
    if hparams is None:
        hparams = PoeemHparam()
    self._warmup_steps = warmup_steps
    self._buffer_size = buffer_size
    self._hparams = hparams
    self._mode = mode
    self._d, self._K, self._D, self._coarse_K, self._rotate = (
      emb_size, hparams.K, hparams.D, hparams.coarse_K, hparams.rotate)
    self._quantizer = PoeemQuantizer(self._coarse_K, self._K, self._D, self._d,
      self._rotate, buffer_size, svd_steps=svd_steps)

  def forward(self, x):
    def forward_layer(x):
      n_batch = tf.shape(x)[0]
      x_tau, coarse_code, code = (
          # the existence of this op is necessary to enable control dependencies working
          tf.identity(x),
          tf.zeros([n_batch], dtype=tf.int64),
          tf.zeros([n_batch, self._D], dtype=tf.int64))
      regularizer = 0.0
      return x_tau, coarse_code, code, regularizer

    def pq_layer(x):
      def accumulate_layer(x):
        with tf.device('/device:CPU:0') if self._hparams.kmeans_on_cpu else (
            contextlib.suppress()):
          deps = self._quantizer.accumulate(x)
          with tf.control_dependencies(deps):
            return forward_layer(x)

      def init_layer(x):
        with tf.device('/device:CPU:0') if self._hparams.kmeans_on_cpu else (
            contextlib.suppress()):
          deps = self._quantizer.init_centroids(
            max_iter=self._hparams.kmeans_max_iter,
            change_pct_thr=self._hparams.kmeans_change_pct_thr)
          with tf.control_dependencies(deps):
            clear_deps = self._quantizer.clear_sample_buffer()
            with tf.control_dependencies(clear_deps):
              return internal_pq_layer(x, init=True)

      def internal_pq_layer(x, init=False):
        rotation_matrix=self._quantizer._rotation_matrix
        if init and rotation_matrix is not None:
          rotation_matrix = tf.stop_gradient(rotation_matrix)
        x_tau, coarse_code, code = self._quantizer.forward(x, rotation_matrix)
        regularizer = tf.reduce_sum((x_tau - tf.stop_gradient(x))**2)
        if coarse_code is None: # dummy placeholder for tf.cond.
          coarse_code = tf.zeros([tf.shape(x)[0]], dtype=tf.int64)
        return x_tau, coarse_code, code, regularizer

      if self._quantizer._buffer_size < 0 or self._mode != ModeKeys.TRAIN:
        return internal_pq_layer(x) # skip kmeans/svd initialization.
      return tf.cond(tf.less(self._quantizer._sample_size, self._quantizer._buffer_size),
        lambda: accumulate_layer(x),    # accumulate data
        lambda: tf.cond(
          tf.equal(self._quantizer._initialized, 0),
          lambda: init_layer(x),        # init centroids
          lambda: internal_pq_layer(x)))

    if self._mode == ModeKeys.TRAIN:
      # forward_layer: 1. steps from 0 to _warmup_steps, train model without pq
      # pq_layer: 2. steps from _warmup_steps to _buffer_size/batch_size,
      #              train model without pq and accumulate input data in _sample_buffer
      #           3. clustering input data in _sample_buffer to init centroids and start to train pq model
      step = tf.cast(tf.train.get_global_step(), tf.int32)
      if self._warmup_steps > 0:
        x_tau, coarse_code, code, regularizer = tf.cond(
          tf.less_equal(step, self._warmup_steps),
          true_fn=lambda: forward_layer(x),
          false_fn=lambda: pq_layer(x))
    else:
      x_tau, coarse_code, code, regularizer = pq_layer(x)

    update_ops = self.metrics(x, x_tau, coarse_code, code)
    with tf.control_dependencies(update_ops):
      x_tau = tf.identity(x_tau)
    return x_tau, coarse_code, code, regularizer

  def metrics(self, x, x_tau, coarse_code, code):
    distortion = compute_distortion(x, x_tau)
    tf.summary.scalar('distortion', distortion)
    update_ops = []
    def full_summary(dim, code, name):
      v, idx, counts = tf.unique_with_counts(code)
      tf.summary.scalar(
        'code_distribution/unique_%s_count' % name, tf.shape(v)[0])
      tf.summary.histogram(
        'code_distribution/%s_histogram' % name, code)
      tally_var = tf.Variable(tf.zeros([dim], dtype=tf.int64),
        name=name + '_cumulative_count', dtype=tf.int64, trainable=False,
        aggregation=tf.VariableAggregation.SUM)
      update_ops.append(tf.scatter_add(tally_var, v, tf.to_int64(counts)))
      tally_var = tf.to_float(tally_var)
      cum_frac = (tally_var + 1e-8) / (tf.reduce_sum(tally_var) + 1e-8 * dim)
      kl = tf.keras.losses.KLDivergence(reduction=tf.compat.v1.losses.Reduction.NONE)
      # ideally cluster sizes should be uniform, so track distances to uniform.
      uniform = tf.ones([dim], dtype=tf.float32) / dim
      def tv(dist1, dist2): # total variation distance of two prob. measures.
        return tf.reduce_sum(tf.abs(dist1 - dist2)) / 2.0
      tf.summary.scalar('code_distribution/%s_kld' % name, kl(cum_frac, uniform))
      tf.summary.scalar('code_distribution/%s_tvd' % name, tv(cum_frac, uniform))
    if self._coarse_K > 0:
      full_summary(self._coarse_K, coarse_code, 'coarse_code')
    code_part0 = code[:, 0]
    full_summary(self._K, code_part0, 'code0')
    return update_ops


class PoeemHparam(object):
  # A default Poeem parameter setting (demo)
  def __repr__(self):
    return str({k: getattr(self, k) for k in dir(self) if
        not k.startswith('_') and 'built-in' not in str(getattr(self, k))})

  def __init__(self,
               coarse_K=128,
               K=16,
               D=32,
               rotate=0,
               kmeans_max_iter=100,
               kmeans_change_pct_thr=1e-2,
               kmeans_on_cpu=False):
    """
    Args:

    """
    self.coarse_K = coarse_K
    self.K = K
    self.D = D
    self.rotate = rotate
    self.kmeans_max_iter = kmeans_max_iter
    self.kmeans_change_pct_thr = kmeans_change_pct_thr
    self.kmeans_on_cpu = kmeans_on_cpu # gpu may run OOM.



import tensorflow as tf
    
def encode(data, codebook):
    """Encode given item embeddings into K-way D-group code with codebook
    
    For each d/D length subvector, find the nearest vector in the sub-codebook, 
    take the index of the nearest vector as the encoded value

    This function is supposed to be used to export item embeddings

    Args:
        data: a tensor with [n_batch, d]
        codebook: a tensor with [D, K, d/D]

    Returns:
        code: a tensor with [n_batch, D]
    """
    n_item, d = tf.shape(data)[0], tf.shape(data)[1]
    D, K, dd = tf.shape(codebook)[0], tf.shape(codebook)[1], tf.shape(codebook)[2]
    with tf.control_dependencies([tf.assert_equal(dd * D, d, message='inputs and codebook must have consistent dimensions')]):
        reshaped_data = tf.reshape(data, [n_item, D, dd])  # (n_batch, D, dd)
        norm_1 = tf.reduce_sum(reshaped_data**2, -1, keep_dims=True)  # (n_batch, D, 1)
        norm_2 = tf.expand_dims(tf.reduce_sum(codebook**2, -1), 0)  # (1, D, K)
        dot = tf.matmul(tf.transpose(reshaped_data, perm=[1, 0, 2]),
                        tf.transpose(codebook, perm=[0, 2, 1]))  # (D, n_batch, K)
        distance = norm_1 - 2 * tf.transpose(dot, perm=[1, 0, 2]) + norm_2
        distance = tf.reshape(distance, [-1, D, K])  # (n_batch, D, K)
        code = tf.argmin(distance, axis=-1)  # (n_batch, D)
        return code



import tensorflow as tf
from tensorflow.python.training.tracking import base as trackable_base
from tensorflow.python.training.tracking import tracking as trackable

from poeem.ops.bin import search_op

def index_from_file(index_file, name=None):
    if index_file is None or len(index_file) == 0:
        raise ValueError("index_file must be specified and must not be empty.")

    with tf.name_scope(name, "string_to_index"):
        with tf.name_scope(None, "index"):
            init = IndexInitializer(index_file)
            index = Index(init)
    return index


class IndexInitializer(trackable_base.Trackable):

    def __init__(self, filename, name=None):
        self._name = name
        self._filename_arg = filename
        self._filename = self._track_trackable(
        trackable.TrackableAsset(filename), "_filename")
    
    @property
    def _shared_name(self):
        shared_name = "index_%s" % (self._filename_arg)
    
    def initialize(self, index):
        with tf.name_scope(self._name, "index_file_init", (index.resource_handle,)):
            filename = tf.convert_to_tensor(
                self._filename, tf.string, name="asset_filepath")
            init_op = search_op.initialize_index_from_file(
                index.resource_handle, filename)
        tf.add_to_collection(tf.GraphKeys.TABLE_INITIALIZERS, init_op)
        tf.add_to_collection(tf.GraphKeys.ASSET_FILEPATHS, filename)
        return init_op


class Index(trackable.TrackableResource):
    def __init__(self, initializer):
        super(Index, self).__init__()
        if isinstance(initializer, trackable_base.Trackable):
            self._initializer = self._track_trackable(initializer, "_initializer")
        with tf.init_scope():
            self._resource_handle = self._create_resource()
            self._init_op = self._initialize()
        
    def _initialize(self):
        return self._initializer.initialize(self)

    def _create_resource(self):
        index_ref = search_op.index()
        self._index_name = index_ref.name.split("/")[-1]
        return index_ref
    
    @property
    def name(self):
        return self._index_name
    
    @property
    def initializer(self):
        return self._init_op

    def search(self, query, n_neighbor, n_probe, metric_type, verbose=True, name=None):
        neighbors, scores = search_op.index_search(
            self.resource_handle, query, n_neighbor, n_probe, metric_type)
        return neighbors, scores



import tensorflow as tf
from tensorflow.python.framework import ops
from tensorflow.python.data.ops import dataset_ops
from tensorflow.python.framework import dtypes

from poeem.ops.bin import pqknn_dataset_op


class PQKnnDataset(dataset_ops.Dataset):
    def _apply_options(self):   # otherwise tf1.15 would apply OptimizeDataset.
        return self

    def with_options(self, options):    # needed for MirroredStrategy.
        del options
        return self

    def __init__(self,
                 batch_file='',
                 item_feature_dict_file='',
                 positive_item_column_index = -1,
                 random_negative_item_count = 0):

        # sanity check
        assert len(batch_file) > 0, 'batch_file must be non-empty'
        assert len(item_feature_dict_file) > 0, 'item_feature_dict_file must be non-empty'

        self._batch_file = ops.convert_to_tensor(
            batch_file,
            dtype=dtypes.string,
            name="batch_file")
        self._item_feature_dict_file = ops.convert_to_tensor(
            item_feature_dict_file,
            dtype=dtypes.string,
            name="item_feature_dict_file")
        self._positive_item_column_index = positive_item_column_index
        self._random_negative_item_count = random_negative_item_count
        # Create any input attrs or tensors as members of this class.
        super(PQKnnDataset, self).__init__()

    def _as_variant_tensor(self):
        return pqknn_dataset_op.pqknn_dataset(
            self._batch_file,
            self._item_feature_dict_file,
            self._positive_item_column_index,
            self._random_negative_item_count)

    def _inputs(self):
        return []

    @property
    def output_types(self):
        return tf.string

    @property
    def output_shapes(self):
        return tf.TensorShape([])

    @property
    def output_classes(self):
        return tf.Tensor



import tensorflow as tf
from poeem.ops.bin import clustering_op
from poeem.ops.bin import clustering_raw_op


def kmeans(code, codebook, n_cluster, sample_size, max_iter=100, change_percentage_thr=0.0001):
    """Do clustering for given item embeddings into K clusters

    Args:
        code: a tensor with [n_batch, D]
        codebook: a tensor with [D, K, d/D]
        n_cluster: a scalar indicates how many clusters

    Returns:
        centroid: a tensor with [K, d]
        assignment: a tensor with [n_batch]
    """
    centroid, assignment = clustering_op.clustering(
        code, codebook, n_cluster, sample_size, max_iter, change_percentage_thr)
    return centroid, assignment


def kmeans_raw(data, n_cluster, max_iter=100, change_percentage_thr=0.0001, verbose=2):
    """Do clustering for given item embeddings into K clusters

    Args:
        data: a tensor with [n_batch, d]
        n_cluster: a scalar indicates how many clusters

    Returns:
        centroid: a tensor with [K, d]
        assignment: a tensor with [n_batch]
    """
    centroid, assignment = clustering_raw_op.clustering_raw(
        data, n_cluster, max_iter, change_percentage_thr, verbose=verbose)
    return centroid, assignment



import tensorflow as tf
from poeem.ops.bin import combinatorial_op

# this is adpated from tf.contrib.images.bipartite_match to be undirected.
def bipartite_match(distance_mat):
    """
    Args:
        distance_mat: an nxn matrix of distances.

    Returns:
        - indices matched to each item, e.g., [1, 0, 3, 2] means the match is (0, 1), (2, 3).
    """
    return combinatorial_op.undirected_bipartite_match(distance_mat, -1.0)    # -1.0 means all rows.


if __name__ == '__main__':
    print(tf.Session().run(bipartite_match([[1, 2], [3, 4]])))



import tensorflow as tf
from tensorflow.python.framework import ops
from tensorflow.python.data.ops import dataset_ops
from tensorflow.python.framework import dtypes

from poeem.ops.bin import rec_dataset_op

class RecDataset(dataset_ops.Dataset):
    def _apply_options(self):   # otherwise tf1.15 would apply OptimizeDataset.
        return self

    def __init__(self,
                 train_file = "",
                 user_file = "",
                 item_file = "",
                 user_column_index = 0,
                 item_column_index = 1,
                 neg_item_count = 0):
        self.train_file = ops.convert_to_tensor(
            train_file,
            dtype=dtypes.string,
            name="train_file")
        self.user_file = ops.convert_to_tensor(
            user_file,
            dtype=dtypes.string,
            name="user_file")
        self.item_file = ops.convert_to_tensor(
            item_file,
            dtype=dtypes.string,
            name="item_file")
        self.user_column_index = ops.convert_to_tensor(
            user_column_index,
            dtype=dtypes.int32,
            name="user_column_index")
        self.item_column_index = ops.convert_to_tensor(
            item_column_index,
            dtype=dtypes.int32,
            name="item_column_index")
        self.neg_item_count = ops.convert_to_tensor(
            neg_item_count,
            dtype=dtypes.int32,
            name="neg_item_count")

        # Create any input attrs or tensors as members of this class.
        super(RecDataset, self).__init__()

    def _as_variant_tensor(self):
        return rec_dataset_op.rec_dataset(
            self.train_file,
            self.user_file,
            self.item_file,
            self.user_column_index,
            self.item_column_index,
            self.neg_item_count)

    def _inputs(self):
      return []

    @property
    def output_types(self):
        return tf.string

    @property
    def output_shapes(self):
        return tf.TensorShape([])

    @property
    def output_classes(self):
        return tf.Tensor



import tensorflow as tf

from poeem.ops.bin import tokenizer_op

def unigrams_and_en_trigram_parser(query):
    """
    Args:
        query: a tensor with [n_batch, ]

    Returns:
        sparse tensor
    """
    return tf.SparseTensor(*tokenizer_op.unigrams_and_en_trigram_parser(query))

def bigrams_and_en_trigram_parser(query):
    """
    Args:
        query: a tensor with [n_batch, ]

    Returns:
        sparse tensor
    """
    return tf.SparseTensor(*tokenizer_op.bigrams_and_en_trigram_parser(query))
